import os, string, re, math
import numarray
import numarray.mlab
from numarray import *
import numarray.linear_algebra as la
from sets import Set



english_stop_words = ["","a","about","above","across","after","against","along","always","am","among","an","any","and","another","anybody","anyone","anything","are","around","as","at","be","before","behind","below","beside","besides","between","beyond","but","by","concerning","do","down","during","each","either","every","except","few","for","from","have","he","her","hers","herself","himself","his","i","in","inside","into","is","it","its","itself","like","many","mine","much","my","myself","near","neither","no","nobody","none","nor","nothing","notwithstanding","of","off","on","one","oneself","opposite","or","other","our","ours","ourselves","outside","over","per","round","she","since","some","somebody","someone","something","that","the","these","they","their","theirs","themselves","this","those","throgh","throughout","till","to","towards","under","until","up","was","we","were","what","which","who","whose","with","within","without","you","your","tours","yourself","yourselves"]


# -------------------------------------------------------------------
""" Weight functions taken from this paper:
Preslav Nakov, Antonia Popova, Plamen Mateev, "Weight functions impact on LSA performance" http://www.cs.berkeley.edu/~nakov/selected_papers_list/preslav_RANLP2001.pdf
"""

# return the frequency (ie, do nothing)
def LWF_frequency(matrix,i,j):
	return (matrix[i][j])

# return the log of the (freq + 1)
def LWF_logarithm(matrix,i,j):
	return math.log(matrix[i][j]+1)

def GWF_normal(matrix,i):
	# get the row
	row =  matrix[i,:]	
	# sum the square of word frequencies 
	total = reduce( lambda x,y : x+y ,[(wf*wf) for wf in row])
	# return final answer
	return 1.0 / math.sqrt(total)

# -------------------------------------------------------------------


def export_to_csv(matrix,filename):
	f = open(filename,"w")
	(rows,cols) = matrix.getshape()
	for i in range(rows):
		for j in range(cols):
			f.write( repr(matrix[i,j]) )
			if (j<cols-1):
				f.write(",")
		f.write("\n")
	f.close()

# -------------------------------------------------------------------
class lsa_model:

	def load_strings (self, document_strings, stop_words=english_stop_words):
		
		# Read a document, return dictionary of word frequencies
		def get_freq_dict(document_string):

			document_string=string.lower(document_string)
			document_string = re.sub("[ \n]+", " ", document_string)
			document_string = re.sub("[^a-z\n ]+", "", document_string)

			# split into a list
			words = string.split( document_string," ")

			word_dictionary = {}
			for word in words:
				word = string.strip(word)
				if word_dictionary.has_key(word):
					word_dictionary[word] = word_dictionary[word] + 1
				else:
					word_dictionary[word] = 1

			# clean out stop words
			for word in english_stop_words:
				if word_dictionary.has_key(word):
					print "removed stop word: " + word
					del word_dictionary[word]
					
			return word_dictionary
		
		# read files and compute word frequency data
		documents_word_freq_dict = {}
		for document_string in document_strings:
			documents_word_freq_dict[document_string] = get_freq_dict(document_string);
		
		self.document_columns = document_strings

		self.make_matrix(documents_word_freq_dict)

	
	def load_files (self, files, stop_words=english_stop_words):

		# Read a document, return dictionary of word frequencies
		def read_document(filepath):

			if not os.path.exists(filepath):
				raise "LSA: specified file %s does not exist" % (filepath)

			# read the file
			f=open(filepath, 'r')
			contents=f.read()
			contents=string.lower(contents)
			contents = re.sub("[ \n]+", " ", contents)
			contents = re.sub("[^a-z\n ]+", "", contents)
			f.close()

			# split into a list
			words = string.split( contents," ")

			word_dictionary = {}
			for word in words:
				word = string.strip(word)
				if word=='':
					continue
				if word_dictionary.has_key(word):
					word_dictionary[word] = word_dictionary[word] + 1
				else:
					word_dictionary[word] = 1

			# clean out stop words
			for word in english_stop_words:
				if word_dictionary.has_key(word):
					print "removed stop word: " + word
					del word_dictionary[word]

			return word_dictionary

		# read files and compute word frequency data
		documents_word_freq_dict = {}
		for file in files:
			documents_word_freq_dict[file] = read_document(file);
			
		self.document_columns = files
		self.make_matrix(documents_word_freq_dict)
		
		
	def make_matrix(self,documents_word_freq_dict,vartype=Float64):
		
		# compile a master word list containing all words used in all documents
		master_word_set = Set([])
		for document in documents_word_freq_dict.keys():		
			master_word_set = master_word_set.union(documents_word_freq_dict[document])
		
		# create empty master array
		master_array = resize( numarray.array([0],type=vartype), (len(master_word_set),len(documents_word_freq_dict.keys()) ) )
		
		# sort words in master word set (not needed, but nice for viewing)
		master_word_list = []
		for word in master_word_set:
			master_word_list.append(word)
		master_word_list.sort()
			
		# make a dictionary containing words matched to their column number
		# (use a dictionary because, using a hash, it has a faster lookup time than a list...or I think it does)
		master_word_dict={}
		row=0
		for key in master_word_list:
			master_word_dict[key] = row
			row=row+1
		
		# fill in the master array [matrix]
		document_count = 0
		for document in self.document_columns:				
			for word in documents_word_freq_dict[document]: 
				master_array[master_word_dict[word]][document_count] = documents_word_freq_dict[document][word]
			document_count=document_count+1

		self.word_rows = master_word_list
		self.matrix = master_array
				
	# find and remove words occuring in only one doc
	def strip_orphans(self):
		(rows,cols) = self.matrix.getshape()	
		strip_row_list = []
		for i in range(rows):
			row = [item for item in self.matrix[i]] 
			# e.g. if all entries are zero except for one...
			if (row.count(0) >= (len(row)-1)):
				strip_row_list.append(i)
				
				# for debug
				print "removed orphan word: "+self.word_rows[i]+ " only in "+self.document_columns[row.index(max(row))]
		
		# now we have a list of rows to be taken out, so take them out
		new_array = resize( 
			numarray.array([0],type=self.matrix.type()), 
			(rows-len(strip_row_list),cols))
		new_word_list = []
		new_array_i = 0
		for i in range(rows):
			# todo: rather than searching the list each time, we should pop the values off (we know they are in order, right?)
			if strip_row_list.count(i)==0:
				new_array[new_array_i] = self.matrix[i] # copy matrix row
				new_word_list.append(self.word_rows[i]) # copy word
				new_array_i=new_array_i+1
		self.word_rows = new_word_list
		self.matrix = new_array

	""" 
	The concept of weights is very well explained in 	
	Preslav Nakov, Antonia Popova, Plamen Mateev, "Weight functions impact on LSA performance" http://www.cs.berkeley.edu/~nakov/selected_papers_list/preslav_RANLP2001.pdf
	"""
	def apply_weights(self, global_weight_function=GWF_normal, local_weight_function=LWF_logarithm):
		(rows,cols) = self.matrix.getshape()
		for i in range(rows):
			global_weight = global_weight_function(self.matrix,i)
			for j in range(cols):
				self.matrix[i][j] = local_weight_function(self.matrix,i,j) * global_weight
				
	def reduce_dimensions(self,new_dimensions):
		[W,SV,P] = numarray.mlab.svd(self.matrix)
		old_dimensions = len(SV)
		for i in range (new_dimensions,old_dimensions):
			SV[i] = 0
		# transform the vector SV into a matrix with SV as the diagonal
		S=SV*identity(len(SV))
		q = matrixmultiply(W,S)
		self.matrix = matrixmultiply(q,P)
	
	#------------------------------------------------------------------
	
	def pretty_print(self):
		# get size of matrix
		(rows,cols) = self.matrix.getshape()
		# first row
		print "%8.8s      " % (" "),
		for document in self.document_columns:
			print "%8.8s " % (document),
		print
		print "%8.8s      " % (" "),
		for col in range(0,len(self.document_columns)):
			print "%8d " % (col),
		print
		for row in range(rows):
			print "%-8.8s %4d|" % (self.word_rows[row],row),
			for col in range(cols):
				print "% 5.5f " % (self.matrix[row][col]),
			print
			
			
	#------------------------------------------------------------------			
	
	def compare_documents(self,compare_function):
		# get size of matrix
		(rows,cols) = self.matrix.getshape()
		# make document comparison matrix
		doc_compare_array = resize( numarray.array([0],type=self.matrix.type()), (cols,cols) )
		for i in range(cols):
			for i2 in range(cols):
				# compare document_i with document_i2
				# (but skip if i2 has already been compared with i)
				#if doc_compare_array[i2][i] != 0:
				#doc_compare_array[i][i2] = compare_function( self.matrix[i,:], self.matrix[i2,:] )
				
				# TODO: this is a hack to convert the numarray into a normal list so stats.lspearmanr can work
				ii = [x for x in self.matrix[i,:]]
				ii2 = [x for x in self.matrix[i2,:]]
				doc_compare_array[i][i2] = compare_function( ii, ii2 )
		return doc_compare_array
				
	def compare_words(self,compare_function):
		# get size of matrix
		(rows,cols) = self.matrix.getshape()
		# make word comparison matrix
		word_compare_array = resize( numarray.array([0],type=self.matrix.type()), (rows,rows) )
		for j in range(rows):
			for j2 in range(rows):
				# compare word _j with word _j2
				# (but skip if j2 has already been compared with j)
				#if doc_compare_array[j2][j] != 0:
				#	doc_compare_array[j][j2] = compare_function( self.matrix[:,j], self.matrix[:,j2] )
				
				# TODO: this is a hack to convert the numarray into a normal list so stats.lspearmanr can work
				jj = [x for x in self.matrix[j,:]]
				jj2 = [x for x in self.matrix[j2,:]]
				word_compare_array[j][j2] = compare_function( jj, jj2 )
		return word_compare_array
				
